from typing import List, Dict

from ..ComponentEntitlementInfo import TokenUsage, SessionUsage, CurrencyUsage, ComponentEntitlementInfo


class JSONComponentEntitlement:
    __fields = ["id", "components", "tokens_entitled", "tokens_in_use", "tokens_in_use_by_session",
                  "tokens_checked_out", "sessions_entitled", "sessions_in_use", "in_use_by_session",
                  "currency_entitled", "currency_used"]

    def __init__(self, values: Dict[str, object]):
        self.id = 0
        self.components = []
        self.tokens_entitled = 0
        self.tokens_in_use = 0
        self.tokens_in_use_by_session = 0
        self.sessions_entitled = 0
        self.sessions_in_use = 0
        self.currency_entitled = 0
        self.currency_used = 0
        self.in_use_by_session = False
        [self.__setattr__(a, values[a]) for a in self.__class__.__fields if a in values]

    def Deserialize(self) -> ComponentEntitlementInfo:
        info = ComponentEntitlementInfo()
        info.ID = self.id
        info.Components = self.components
        info.TokenUsage = TokenUsage(self.tokens_entitled, self.tokens_in_use, self.tokens_in_use_by_session)
        info.SessionUsage = SessionUsage(self.sessions_entitled, self.sessions_in_use, self.in_use_by_session)
        info.CurrencyUsage = CurrencyUsage(self.currency_entitled, self.currency_used)
        return info



from datetime import timedelta
from typing import Optional, Tuple, List, Callable
from uuid import UUID

from .enumerations import ValidationResult
from .offline_session_info_helper import OfflineSessionInfoHelper
from .. import ControlManagerHelper
from ..ComponentCheckoutResult import ComponentCheckoutResult
from ..ComponentInfo import ComponentInfo
from ..ComponentsStatus import ComponentsStatus
from ..InstanceIdentity import InstanceIdentity
from ..SessionValidity import SessionValidity
from ..UserIdentity import UserIdentity
from ..control_logic import ControlLogic
from ..enumerations.licensing_enumerations import *
from ..states.ControlState import ControlState
from ..states.ValidatedSessionState import ValidatedSessionState
from ..states.ValidationState import ValidationState
from ..timehandling import TimeProvider
from ..timehandling.UtcDateTime import UtcDateTime
from ..webservice.LicensingWebService import LicensingWebService
from ..webservice.server_results import BeginAppSessionResult, ValidatedSessionResult, CheckOutComponentResult


class AppSessionManager(object):
    _webService: LicensingWebService
    _productId: UUID
    _version: str
    _validationState: ValidationState
    _sessionState: ValidatedSessionState
    _instanceIdentity: InstanceIdentity
    _userIdentity: UserIdentity

    def __init__(self, web_service: LicensingWebService, product_id: UUID, version: str,
                 validation_state: ValidationState, validated_session_state: ValidatedSessionState,
                 instance_identity: Optional[InstanceIdentity] = None,
                 user_identity: Optional[UserIdentity] = None):
        self._webService = web_service
        self._productId = product_id
        self._version = version
        self._validationState = validation_state
        self._sessionState = validated_session_state
        self._instanceIdentity = instance_identity
        self._userIdentity = user_identity

    @staticmethod
    def Now() -> UtcDateTime:
        return TimeProvider.Current().UtcNow()

    def BeginSession(self, requested_session_duration: Optional[timedelta] = None):
        if self._sessionState.SessionState != SessionState.NotStarted:
            raise RuntimeError()

        web_result = self.BeginWebSession(requested_session_duration)
        validity, result = self.ProcessWebResultIntoValidity(web_result)
        if result is ValidationResult.OnlineSuccess:
            self._sessionState.SessionState = SessionState.Active
        elif result is ValidationResult.OfflineSuccess:
            self._sessionState.SessionState = SessionState.ActiveOffline
            OfflineSessionInfoHelper.BeginOfflineSession(self._validationState, self._sessionState, self.Now(), None)
        return validity

    def BeginWebSession(self, requested_session_duration: Optional[timedelta] = None) -> BeginAppSessionResult:
        result = self._webService.BeginAppSessionAsync(self._productId, self._validationState.InstanceId,
                                                       self._version, self._validationState.UserId,
                                                       self._instanceIdentity, self._userIdentity,
                                                       requested_session_duration)
        if result.Status is ValidationStatus.Success:
            self.UpdateCommonSessionState(result)
            self._sessionState.SessionKey = result.SessionKey
            self._validationState.Components.clear()
            [self._validationState.Components.append(x) for x in result.Components]
            self._validationState.ComponentsLoaded = True
            self._validationState.ComponentEntitlements.clear()
            [self._validationState.ComponentEntitlements.append(x) for x in result.ComponentEntitlements]
            self._validationState.ComponentEntitlementsLoaded = True
        return result

    def UpdateCommonSessionState(self, result: ValidatedSessionResult):
        self._sessionState.AuthToken = result.AuthToken
        self._validationState.GracePeriodForValidationFailures = result.GracePeriodForValidationFailures
        self._validationState.SessionDuration = result.SessionDuration
        features = ControlState.Features.__get__(self._validationState)
        features.clear()
        [features.append(x) for x in result.Features]

    def ProcessWebResultIntoValidity(self, web_result: ValidatedSessionResult) -> \
            Tuple[SessionValidity, ValidationResult]:
        if (web_result.Status == ValidationStatus.ServiceUnreachable or
                web_result.Status == ValidationStatus.GeneralError):
            ControlManagerHelper.LogServerNotReached(self._validationState, self.Now())
            status, expiration = ControlManagerHelper.CheckOfflineStatus(self._validationState, self.Now())
            if status is ControlManagerHelper.OfflineGracePeriodStatus.NotPermitted:
                return SessionValidity.Invalid(InvalidReason.ServiceUnreachable,
                                               False), ValidationResult.OfflineFailure
            elif status is ControlManagerHelper.OfflineGracePeriodStatus.UserTamperingDetected:
                return SessionValidity.Invalid(InvalidReason.UserTamperingDetected,
                                               False), ValidationResult.OfflineFailureUserTampering
            else:
                return SessionValidity.ValidationFailureGracePeriod(expiration), ValidationResult.OfflineSuccess
        else:
            ControlManagerHelper.LogServerReached(self._validationState, web_result.ServerTime, web_result.Status)
            if web_result.Status == ValidationStatus.Success:
                return SessionValidity.Valid(self._validationState.SessionDuration), ValidationResult.OnlineSuccess
            else:
                return SessionValidity.Invalid(ControlManagerHelper.StatusToReason(web_result.Status),
                                               False), ValidationResult.OnlineFailure

    def ExtendSession(self, requested_extension_duration: timedelta) -> SessionValidity:

        session_state = self._sessionState.SessionState
        if session_state is SessionState.Active or session_state is not SessionState.ActiveOfflineWithOnlineSession:
            web_result = self.ExtendWebSession(requested_extension_duration)
        elif session_state is SessionState.ActiveOffline:
            web_result = self.BeginWebSession(requested_extension_duration)
        else:
            raise ValueError('There is no active session to extend')

        validity, result = self.ProcessWebResultIntoValidity(web_result)
        if result is ValidationResult.OnlineSuccess:
            self._sessionState.SessionState = SessionState.Active
        elif result is ValidationResult.OfflineSuccess:
            if session_state is SessionState.Active:
                OfflineSessionInfoHelper.BeginOfflineSession(self._validationState, self._sessionState, None,
                                                             self._sessionState.SessionKey)
            if session_state is SessionState.Active or session_state is SessionState.ActiveOfflineWithOnlineSession:
                self._sessionState.SessionState = SessionState.ActiveOfflineWithOnlineSession
            else:
                self._sessionState.SessionState = SessionState.ActiveOffline
        else:
            self._sessionState.SessionState = SessionState.Ended
        return validity

    def ExtendWebSession(self, requested_extension_duration: timedelta) -> ValidatedSessionResult:
        result = self._webService.ExtendSession(self._productId, self._sessionState.AuthToken,
                                                requested_extension_duration)
        if result.Status is ValidationStatus.Success:
            self.UpdateCommonSessionState(result)
        return result

    def CheckOutComponents(self, components: List[str]) -> ComponentCheckoutResult:
        return self.CheckOutComponentsCommon(components, lambda: self._webService.CheckOutComponents(self._productId, self._sessionState.SessionKey, components))

    def CheckOutConsumableComponent(self, component: str, token_count: Optional[int]) -> ComponentCheckoutResult:
        return self.CheckOutComponentsCommon([component], lambda: self._webService.CheckOutConsumableComponent(self._productId, self._sessionState.SessionKey, component, token_count))

    def CheckOutComponentsCommon(self, components: List[str], web_result_provider: Callable[[], CheckOutComponentResult]) -> ComponentCheckoutResult:
        session_state = self._sessionState.SessionState
        if session_state is SessionState.NotStarted or \
                session_state is SessionState.Ended:
            raise ValueError('A session is not active')
        if session_state is SessionState.Active:
            web_result = web_result_provider()
            if (web_result.Status == ValidationStatus.ServiceUnreachable or
                    web_result.Status == ValidationStatus.GeneralError):
                failure_reason = ComponentCheckoutFailureReason.ServiceUnreachable if \
                    web_result.Status == ValidationStatus.ServiceUnreachable else ComponentCheckoutFailureReason.Unknown
                return self.CheckOutFreeComponentsOffline(components, failure_reason)
            else:
                if web_result.Status is ValidationStatus.Success:
                    return ComponentCheckoutResult.GetSuccess(False)
                elif web_result.Status is ValidationStatus.NoSeatsAvailable:
                    return ComponentCheckoutResult.GetFailure(ComponentCheckoutFailureReason.InsufficientSessions,
                                                              False)
                elif web_result.Status is ValidationStatus.InsufficientTokens:
                    return ComponentCheckoutResult.GetFailure(ComponentCheckoutFailureReason.InsufficientTokens, False)
                elif web_result.Status is ValidationStatus.InvalidComponent:
                    return ComponentCheckoutResult.GetFailure(ComponentCheckoutFailureReason.InvalidComponent, False)
                else:
                    return ComponentCheckoutResult.GetFailure(ComponentCheckoutFailureReason.Unknown, False)
        elif session_state is SessionState.ActiveOffline or \
                session_state is SessionState.ActiveOfflineWithOnlineSession:
            return self.CheckOutFreeComponentsOffline(components, ComponentCheckoutFailureReason.ServiceUnreachable)
        else:
            raise ValueError('A session is not active')

    def CheckOutFreeComponentsOffline(self, components: List[str],
                                      failure_reason: ComponentCheckoutFailureReason) -> ComponentCheckoutResult:
        if ControlLogic.IsTimeInconsistencyDetected(self._validationState.LastSuccessfulValidationTime,
                                                    self._validationState.FailedValidationTimesUtc, self.Now(), False):
            return ComponentCheckoutResult.GetFailure(ComponentCheckoutFailureReason.UserTamperingDetected, True)

        if self._validationState.LastSuccessfulValidationTime is not None:
            return ComponentCheckoutResult.GetFailure(failure_reason, True)

        if not self._validationState.ComponentsLoaded:
            return ComponentCheckoutResult.GetFailure(failure_reason, True)

        failed = False
        for name in components:
            component_info: ComponentInfo = next(
                (c for c in self._validationState.Components if c.Name.casefold() == name.casefold()), None)
            if component_info is None:
                failed = True
                break

            bad = True
            if component_info.LicenseModel is ComponentLicenseModel.Nothing:
                bad = False
            elif component_info.LicenseModel is ComponentLicenseModel.Token:
                effective_tokens = component_info.TokensRequiredAt(self.Now())
                if effective_tokens == 0:
                    bad = False
            if bad:
                failed = True
                break

        if failed:
            return ComponentCheckoutResult.GetFailure(failure_reason, True)
        else:
            return ComponentCheckoutResult.GetSuccess(True)

    def GetComponentsStatus(self) -> ComponentsStatus:
        if self._sessionState.SessionState is SessionState.NotStarted or \
                self._sessionState.SessionState is SessionState.Ended:
            raise ValueError('A session is not active')
        elif self._sessionState.SessionState is SessionState.ActiveOffline:
            ret = ComponentsStatus()
            ret.Success = False
            ret.Components = self._validationState.ComponentsIfLoaded()
            ret.ComponentEntitlements = self._validationState.ComponentEntitlementsIfLoaded()
            return ret
        else:
            web_result = self._webService.GetComponentStatus(self._productId, self._sessionState.SessionKey)
            if web_result.Success:
                self._validationState.Components.clear()
                [self._validationState.Components.append(x) for x in web_result.Components]
                self._validationState.ComponentsLoaded = True
                self._validationState.ComponentEntitlements.clear()
                [self._validationState.ComponentEntitlements.append(x) for x in web_result.ComponentEntitlements]
                self._validationState.ComponentEntitlementsLoaded = True
                ret = ComponentsStatus()
                ret.Success = False
                ret.Components = self._validationState.Components
                ret.ComponentEntitlements = self._validationState.ComponentEntitlements
                return ret
            else:
                ret = ComponentsStatus()
                ret.Success = False
                ret.Components = self._validationState.ComponentsIfLoaded
                ret.ComponentEntitlements = self._validationState.ComponentEntitlementsIfLoaded
                return ret

    def CurrentOfflineSessionID(self) -> Optional[UUID]:
        if self._sessionState.SessionState is SessionState.ActiveOffline or \
                self._sessionState.SessionState is SessionState.ActiveOfflineWithOnlineSession:
            return self._sessionState.OfflineSessionID
        else:
            return None

    def EndSession(self) -> bool:
        session_state = self._sessionState.SessionState
        if session_state is SessionState.NotStarted or \
                session_state is SessionState.Ended:
            raise ValueError('A session is not active')
        elif session_state is SessionState.Active or session_state is SessionState.ActiveOfflineWithOnlineSession:
            result = self._webService.EndSession(self._productId, self._sessionState.AuthToken)
            if result.Success:
                if session_state is SessionState.ActiveOfflineWithOnlineSession:
                    OfflineSessionInfoHelper.EndOfflineSession(self._validationState, self._sessionState, None)
                self._sessionState.SessionState = SessionState.Ended
                return True
            else:
                return False
        else:
            OfflineSessionInfoHelper.EndOfflineSession(self._validationState, self._sessionState, self.Now())
            self._sessionState.SessionState = SessionState.Ended
            return True



from dataclasses import dataclass
from typing import List


@dataclass(frozen=True)
class TokenUsage:
    tokens_entitled: int
    tokens_in_use: int
    tokens_in_use_by_session: int


@dataclass(frozen=True)
class SessionUsage:
    sessions_entitled: int
    sessions_in_use: int
    in_use_by_session: bool

@dataclass(frozen=True)
class CurrencyUsage:
    currency_entitled: int
    currency_used: int


class ComponentEntitlementInfo:
    _id: int
    _components: List[str]
    _tokenUsage: TokenUsage
    _sessionUsage: SessionUsage
    _currencyUsage: CurrencyUsage

    @property
    def ID(self) -> int:
        return self._id

    @ID.setter
    def ID(self, value: int):
        self._id = value

    @property
    def Components(self) -> List[str]:
        return self._components

    @Components.setter
    def Components(self, value: List[str]):
        self._components = list(value)

    @property
    def TokenUsage(self):
        return self._tokenUsage

    @TokenUsage.setter
    def TokenUsage(self, value: TokenUsage):
        self._tokenUsage = value

    @property
    def SessionUsage(self) -> SessionUsage:
        return self._sessionUsage

    @SessionUsage.setter
    def SessionUsage(self, value: SessionUsage):
        self._sessionUsage = value

    @property
    def CurrencyUsage(self) -> CurrencyUsage:
        return self._currencyUsage

    @CurrencyUsage.setter
    def CurrencyUsage(self, value: CurrencyUsage):
        self._currencyUsage = value


from typing import Union

class RegistrationResult:
    _success: bool
    _errorMessage: Union[str, type(None)]

    def __init__(self, success: bool, error_message: Union[str, type(None)]):
        self._success = success
        self._errorMessage = error_message

    @property
    def Success(self) -> bool:
        return self._success

    @property
    def ErrorMessage(self) -> Union[str, type(None)]:
        return self._errorMessage

    @classmethod
    def GetSuccess(cls):
        return cls(True, None)

    @classmethod
    def GetError(cls, error_message: str):
        return cls(False, error_message)



from datetime import *
from .ControlState import ControlState
from ..webservice import server_results


class SessionControlState(ControlState):
    def __init__(self):
        super().__init__()
        self._sessionDuration = timedelta(hours=0)

    @staticmethod
    def fromOther(source, dest):
        if dest is None:
            dest = SessionControlState()
        dest = ControlState.fromOther(source, dest)
        if dest.Registered:
            dest._sessionDuration = source.SessionDuration
        return dest

    @property
    def SessionDuration(self):
        return self._sessionDuration

    @SessionDuration.setter
    def SessionDuration(self, value: timedelta):
        if value is None:
            raise ValueError("value must not be null")
        elif not isinstance(value, timedelta):
            raise ValueError('value must be a timedelta')
        self._sessionDuration = value

    def AdditionalUpdateAfterRegistrationSuccess(self, result: server_results.RegisterUserResult):
        self._sessionDuration = result.SessionDuration



from datetime import datetime
from .UtcDateTime import UtcDateTime


def Default():
    return SystemTimeProvider()


def Current():
    return SystemTimeProvider()


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class SystemTimeProvider(metaclass=Singleton):

    def UtcNow(self):
        return UtcDateTime(datetime.utcnow())



from uuid import UUID
from datetime import datetime
from ..timehandling.UtcDateTime import UtcDateTime
from typing import Optional, Union, List


class OfflineComponentCheckoutInfo:
    _component: str
    _checkoutTime: datetime
    _tokenCount: int

    def __init__(self, component: str, checkout_time: Union[datetime, UtcDateTime], token_count: int):
        self._component = component
        if checkout_time is datetime:
            self._checkoutTime = checkout_time
        elif checkout_time is UtcDateTime:
            self._checkoutTime = checkout_time.AsDateTime()
        self._tokenCount = token_count

    @property
    def Component(self) -> str:
        return self._component

    @Component.setter
    def Component(self, value: str):
        self._component = value

    @property
    def CheckoutTime(self) -> datetime:
        return self._checkoutTime

    @CheckoutTime.setter
    def CheckoutTime(self, value: datetime):
        self._checkoutTime = value

    @property
    def CheckoutTimeUtc(self) -> UtcDateTime:
        return UtcDateTime(self._checkoutTime)

    @CheckoutTimeUtc.setter
    def CheckoutTimeUtc(self, value: UtcDateTime):
        self._checkoutTime = value.AsDateTime()

    @property
    def TokenCount(self) -> int:
        return self._tokenCount

    @TokenCount.setter
    def TokenCount(self, value: int):
        self._tokenCount = value


class OfflineSessionInfo:
    _localSessionId: UUID
    _sessionKey: Union[str, type(None)]
    _startTime: Optional[datetime]
    _endTime: Optional[datetime]
    _componentCheckouts: List[OfflineComponentCheckoutInfo]
    _uploadAttempts: int

    @property
    def LocalSessionId(self) -> UUID:
        return self._localSessionId

    @LocalSessionId.setter
    def LocalSessionId(self, value: UUID):
        self._localSessionId = value

    @property
    def SessionKey(self) -> Union[str, type(None)]:
        return self._sessionKey

    @SessionKey.setter
    def SessionKey(self, value: Union[str, type(None)]):
        self._sessionKey = value

    @property
    def StartTime(self) -> Optional[datetime]:
        return self._startTime

    @StartTime.setter
    def StartTime(self, value: Optional[datetime]):
        self._startTime = value

    @property
    def UtcStartTime(self) -> Optional[UtcDateTime]:
        return UtcDateTime(self._startTime) if self._startTime is not None else None

    @UtcStartTime.setter
    def UtcStartTime(self, value: Optional[UtcDateTime]):
        self._startTime = value.AsDateTime() if value is not None else None

    @property
    def EndTime(self) -> Optional[datetime]:
        return self._endTime

    @EndTime.setter
    def EndTime(self, value: Optional[datetime]):
        self._endTime = value

    @property
    def UtcEndTime(self) -> Optional[UtcDateTime]:
        return UtcDateTime(self._endTime) if self._endTime is not None else None

    @UtcEndTime.setter
    def UtcEndTime(self, value: Optional[UtcDateTime]):
        self._endTime = value.AsDateTime() if value is not None else None

    @property
    def ComponentCheckouts(self) -> List[OfflineComponentCheckoutInfo]:
        return self._componentCheckouts

    @ComponentCheckouts.setter
    def ComponentCheckouts(self, value: List[OfflineComponentCheckoutInfo]):
        self._componentCheckouts = value

    @property
    def UploadAttempts(self) -> int:
        return self._uploadAttempts

    @UploadAttempts.setter
    def UploadAttempts(self, value: int):
        self._uploadAttempts = value

    def AddComponentCheckout(self, info: OfflineComponentCheckoutInfo):
        if self._componentCheckouts is None:
            self._componentCheckouts = [info]
        else:
            self._componentCheckouts.append(info)



from ..states.ValidationState import ValidationState
from ..states.ValidatedSessionState import ValidatedSessionState
from ..timehandling.UtcDateTime import UtcDateTime
from uuid import UUID, uuid4
from typing import Optional, Union
from .OfflineSessionInfo import OfflineSessionInfo


class OfflineSessionInfoHelper:

    @staticmethod
    def BeginOfflineSession(state: ValidationState, session_state: ValidatedSessionState,
                            start_time: Optional[UtcDateTime], server_session_key: Union[str, type(None)]):
        id = uuid4()
        osi = OfflineSessionInfo()
        osi.LocalSessionId = id
        osi.UtcStartTime = start_time
        osi.SessionKey = server_session_key
        state.OfflineSessionsUnsafe.append(osi)
        session_state.OfflineSessionID = id
        return id

    @staticmethod
    def EndOfflineSession(state: ValidationState, session_state: ValidatedSessionState,
                          end_time: Optional[UtcDateTime]):
        osi = state.OfflineSession(session_state.OfflineSessionID)
        osi.UtcEndTime = end_time



from .enumerations.licensing_enumerations import InvalidReason
from datetime import datetime, timedelta
from .timehandling.UtcDateTime import UtcDateTime
from .timehandling import TimeProvider


class SessionValidity:
    _isValid: bool = False
    _invalidReason: InvalidReason = InvalidReason.Unknown
    _isInValidationFailureGracePeriod: bool = False
    _isInUnregisteredGracePeriod: bool = False
    _isOfflineCheckout: bool = False
    _hasExpiredOfflineCheckout: bool = False
    _gracePeriodExpiration: datetime = datetime.min
    _validityDuration: timedelta = timedelta(days=0)

    @property
    def IsValid(self) -> bool:
        return self._isValid

    @IsValid.setter
    def IsValid(self, value: bool):
        self._isValid = value

    @property
    def InvalidReason(self) -> InvalidReason:
        return self._invalidReason

    @InvalidReason.setter
    def InvalidReason(self, value: InvalidReason):
        self._invalidReason = value

    @property
    def IsInValidationFailureGracePeriod(self) -> bool:
        return self._isInValidationFailureGracePeriod

    @IsInValidationFailureGracePeriod.setter
    def IsInValidationFailureGracePeriod(self, value: bool):
        self._isInValidationFailureGracePeriod = value

    @property
    def IsInUnregisteredGracePeriod(self) -> bool:
        return self._isInUnregisteredGracePeriod

    @IsInUnregisteredGracePeriod.setter
    def IsInUnregisteredGracePeriod(self, value: bool):
        self._isInUnregisteredGracePeriod = value

    @property
    def IsOfflineCheckout(self) -> bool:
        return self._isOfflineCheckout

    @IsOfflineCheckout.setter
    def IsOfflineCheckout(self, value: bool):
        self._isOfflineCheckout = value

    @property
    def HasExpiredOfflineCheckout(self) -> bool:
        return self._hasExpiredOfflineCheckout

    @HasExpiredOfflineCheckout.setter
    def HasExpiredOfflineCheckout(self, value: bool):
        self._hasExpiredOfflineCheckout = value

    @property
    def GracePeriodExpiration(self) -> datetime:
        return self._gracePeriodExpiration

    @GracePeriodExpiration.setter
    def GracePeriodExpiration(self, value: datetime):
        self._gracePeriodExpiration = value

    @property
    def ValidityDuration(self) -> timedelta:
        return self._validityDuration

    @ValidityDuration.setter
    def ValidityDuration(self, value: timedelta):
        self._validityDuration = value

    @classmethod
    def Invalid(cls, reason: InvalidReason, has_expired_offline_checkout: bool):
        ret = cls()
        ret.IsValid = False
        ret.InvalidReason = reason
        ret.HasExpiredOfflineCheckout = has_expired_offline_checkout
        return ret

    @classmethod
    def Valid(cls, validity_duration: timedelta):
        ret = cls()
        ret.IsValid = True
        ret.ValidityDuration = validity_duration
        return ret

    @classmethod
    def UnregisteredGracePeriod(cls, expiration: UtcDateTime):
        ret = cls()
        ret.IsValid = True
        ret.ValidityDuration = expiration - TimeProvider.Current().UtcNow()
        ret.IsInUnregisteredGracePeriod = True
        ret.GracePeriodExpiration = expiration.AsDateTime()
        return ret

    @classmethod
    def ValidationFailureGracePeriod(cls, expiration: UtcDateTime):
        ret = cls()
        ret.IsValid = True
        ret.ValidityDuration = expiration - TimeProvider.Current().UtcNow()
        ret.IsInValidationFailureGracePeriod = True
        ret.GracePeriodExpiration = expiration.AsDateTime()
        return ret

    @classmethod
    def OfflineCheckout(cls, expiration: UtcDateTime):
        ret = cls()
        ret.IsValid = True
        ret.ValidityDuration = expiration - TimeProvider.Current().UtcNow()
        ret.IsOfflineCheckout = True
        return ret



from datetime import datetime, tzinfo, timedelta
from typing import Dict
import pytz

from ..ComponentInfo import ComponentInfo
from ..enumerations.licensing_enumerations import ComponentLicenseModel

SERVER_TIME_FORMAT = "%a, %d %b %Y %H:%M:%S %z"


class JSONComponent:
    __fields = ["name", "license_model", "tokens_required", "free_trial", "free_trial_expiration",
              "original_tokens_required"]

    def __init__(self, values: Dict[str, object]):
        self.name = ""
        self.license_model = ""
        self.tokens_required = 0
        self.free_trial = False
        self.free_trial_expiration = datetime.min.replace(tzinfo=pytz.utc).strftime(SERVER_TIME_FORMAT)
        self.original_tokens_required = 0
        [self.__setattr__(a, values[a]) for a in self.__class__.__fields if a in values]

    def Deserialize(self) -> ComponentInfo:
        info = ComponentInfo()
        info.Name = self.name
        info.LicenseModel = ComponentLicenseModel.GetValue(self.license_model)
        info.TokensRequired = self.tokens_required
        info.HasFreeTrial = self.free_trial
        info.FreeTrialExpirationTime = datetime.strptime(self.free_trial_expiration, SERVER_TIME_FORMAT)
        info.OriginalTokensRequired = self.original_tokens_required
        return info



from typing import List
from .ComponentInfo import ComponentInfo
from .ComponentEntitlementInfo import ComponentEntitlementInfo


class ComponentsStatus:
    _success: bool
    _components: List[ComponentInfo]
    _componentEntitlements: List[ComponentEntitlementInfo]

    @property
    def Success(self) -> bool:
        return self._success

    @Success.setter
    def Success(self, value: bool):
        self._success = value

    @property
    def Components(self) -> List[ComponentInfo]:
        return self._components

    @Components.setter
    def Components(self, value: List[ComponentInfo]):
        self._components = value

    @property
    def ComponentEntitlements(self) -> List[ComponentEntitlementInfo]:
        return self._componentEntitlements

    @ComponentEntitlements.setter
    def ComponentEntitlements(self, value: List[ComponentEntitlementInfo]):
        self._componentEntitlements = value



from typing import Optional, Dict

DEFAULT_CASE_SENSITIVE: bool = True
DEFAULT_WEIGHT: int = 1
DEFAULT_MAX_ARRAY_VALUE_DIFFERENCE: int = 1
DEFAULT_VERSION = 1
DEFAULT_MAX_MISMATCH_COUNT = 1


class InstanceIdentityValue(object):
    def __init__(self, value: object, case_sensitive: Optional[bool] = DEFAULT_CASE_SENSITIVE,
                 weight: Optional[int] = DEFAULT_WEIGHT,
                 max_array_value_difference: Optional[int] = DEFAULT_MAX_ARRAY_VALUE_DIFFERENCE):
        if value is None:
            raise ValueError('Value cannot be None')
        self._value = value
        self._caseSensitive = case_sensitive
        self._weight = weight
        self._maxArrayValueDifference = max_array_value_difference

    @property
    def Value(self) -> object:
        return self._value

    @property
    def CaseSensitive(self) -> bool:
        return self._caseSensitive

    @property
    def Weight(self) -> int:
        return self._weight

    @property
    def MaxArrayValueDifference(self) -> int:
        return self._maxArrayValueDifference


class InstanceIdentity(object):
    def __init__(self, values: Dict[str, InstanceIdentityValue], version: Optional[int] = DEFAULT_VERSION,
                 maximum_mismatch_count: Optional[int] = DEFAULT_MAX_MISMATCH_COUNT):
        self._values = dict(values)
        self._version = version
        self._maximumMismatchCount = maximum_mismatch_count

    @property
    def Version(self) -> int:
        return self._version

    @property
    def MaximumMismatchCount(self) -> int:
        return self._maximumMismatchCount

    @property
    def Values(self) -> Dict[str, InstanceIdentityValue]:
        return dict(self._values)



from .timehandling.UtcDateTime import UtcDateTime
from . import ControlStrategy
from datetime import timedelta
from typing import Optional, List, Tuple


class ControlLogic:

    @staticmethod
    def IsFirstLaunchTimeInvalid(first_launch_time: UtcDateTime, current_time: UtcDateTime) -> bool:
        return current_time < first_launch_time

    @staticmethod
    def IsTimeInconsistencyDetected(last_trusted_time: Optional[UtcDateTime], previous_times: List[UtcDateTime],
                                    current_time: UtcDateTime, current_time_is_trusted: bool):
        if last_trusted_time is None:
            return False

        if current_time_is_trusted:
            return False

        return current_time < last_trusted_time or any([current_time < t for t in previous_times])

    @staticmethod
    def IsInGracePeriodForUnregisteredProduct(control_strategy: ControlStrategy, first_launch_time: UtcDateTime,
                                              current_time: UtcDateTime) -> Tuple[bool, UtcDateTime]:

        expiration = first_launch_time + control_strategy.GracePeriodForUnregisteredProduct
        return current_time < expiration, expiration

    @staticmethod
    def IsInGracePeriodForValidationFailures(grace_period_for_validation_failures: timedelta,
                                             last_successful_validation_time: UtcDateTime,
                                             current_time: UtcDateTime) -> Tuple[bool, UtcDateTime]:
        expiration = last_successful_validation_time + grace_period_for_validation_failures
        return current_time < expiration, expiration



from typing import Dict


class UserIdentity(object):

    def __init__(self, data: Dict[str, str]):
        self._data = dict()
        if not (data is None):
            self._data.update(data)

    @property
    def Data(self) -> dict:
        return dict(self._data)



from enum import Enum


class ComponentLicenseModel(Enum):
    Nothing = 0
    Token = 1
    Session = 2
    ConsumableToken = 3
    Unknown = 4

    @staticmethod
    def GetValue(value: str):
        try:
            return ComponentLicenseModel[value]
        except KeyError:
            return ComponentLicenseModel["Unknown"]


class SessionState(Enum):
    NotStarted = 0
    Active = 1
    Ended = 2
    # added in 4.0
    ActiveOffline = 3
    ActiveOfflineWithOnlineSession = 4
    # added in 5.0
    ActiveOfflineCheckout = 5


class ValidationStatus(Enum):
    # Returned when registration is required before starting a session.
    RegistrationRequired = -101
    UserTamperingDetected = -100
    ServiceUnreachable = -99

    # Statuses above this line are client-only.
    # --------------
    # Statuses below this line are sent by the server

    OfflineCheckoutNotSupported = -16
    InvalidComponent = -15
    InsufficientTokens = -14
    ActivationPending = -13
    SessionRevoked = -12
    NoSeatsAvailable = -11
    InconsistentRegistration = -10
    InconsistentUser = -9
    InstanceNotRegistered = -8
    UserNotRegistered = -7
    InvalidLicenseKey = -6
    InvalidProductId = -5
    InstanceDisabled = -4
    UserDisabled = -3
    LicenseExpired = -2
    GeneralError = -1
    Success = 0


class InvalidReason(Enum):
    Unknown = -1
    NotInvalid = 0
    UserTamperingDetected = 1
    ServiceUnreachable = 2
    LicenseExpired = 3
    SessionRevoked = 4
    NoSeatsAvailable = 5
    InconsistentRegistration = 6
    InconsistentUser = 7
    InstanceNotRegistered = 8
    UserNotRegistered = 9
    InvalidLicenseKey = 10
    InvalidProductId = 11
    InstanceDisabled = 12
    UserDisabled = 13
    ActivationPending = 14
    OfflineCheckoutExpired = 15
    OfflineCheckoutNotSupported = 16
    InsufficientTokens = 17


class ComponentCheckoutFailureReason(Enum):
    Unknown = -1
    NoFailure = 0
    ServiceUnreachable = 1
    UserTamperingDetected = 2
    InsufficientTokens = 3
    InsufficientSessions = 4
    InvalidComponent = 5



from typing import Union, List

class UserLicenseResult:
    _success: bool
    _errorMessage: Union[str, type(None)]
    _licenses: List[str]

    def __init__(self, success: bool, error_message: Union[str, type(None)], licenses: List[str]):
        self._success = success
        self._errorMessage = error_message
        self._licenses = licenses.copy()

    @property
    def Success(self) -> bool:
        return self._success

    @property
    def ErrorMessage(self) -> Union[str, type(None)]:
        return self._errorMessage

    @property
    def Licenses(self) -> List[str]:
        return self._licenses




from ..enumerations.licensing_enumerations import ValidationStatus
import datetime
from ..timehandling import UtcDateTime
from typing import Optional, Union


class ControlState(object):

    _registeredAt: UtcDateTime
    _registered: bool

    def __init__(self):
        self._failedRegistrationTimes = []
        self._failedValidationTimes = []
        self._features = []
        self._registered = False
        self._registeredAt = None
        self._gracePeriodForValidationFailures = datetime.timedelta()
        self._lastSuccessfulValidationTime = None
        self._firstLaunchTime = None
        self._lastValidationStatus = ValidationStatus.RegistrationRequired

    @staticmethod
    def fromOther(source, dest):
        if dest is None:
            dest = ControlState()
        dest._firstLaunchTime = source.FirstLaunchTime
        if (source.RegisteredAt is None):
            dest._registered = False
            dest._registeredAt = None
            dest._gracePeriodForValidationFailures = datetime.timedelta()
        else:
            dest._registered = True
            dest._registeredAt = UtcDateTime.UtcDateTime(source.RegisteredAt)
            dest._gracePeriodForValidationFailures = source.GracePeriodForValidationFailures

        if not (source.LastSuccessfulValidationTime is None):
            dest._lastSuccessfulValidationTime = UtcDateTime.UtcDateTime(source.LastSuccessfulValidationTime)
        dest._lastValidationStatus = source.LastValidationStatus
        dest._failedRegistrationTimes = source.FailedRegistrationTimes
        dest._failedValidationTimes = source.FailedValidationTimes
        dest._features = source.Features
        return dest

    @property
    def Features(self):
        return self._features

    @property
    def FailedRegistrationTimes(self):
        return list(self._failedRegistrationTimes)

    # Do not use this in consuming applications
    @property
    def FailedRegistrationTimesUnsafe(self):
        """
        Do not use this property in consuming applications
        """
        return self._failedRegistrationTimes

    @property
    def FailedRegistrationTimesUtc(self):
        return [UtcDateTime.UtcDateTime(x) for x in self._failedRegistrationTimes]

    @property
    def FailedValidationTimes(self):
        return list(self._failedValidationTimes)


    @property
    def FailedValidationTimesUnsafe(self):
        """
        Do not use this property in consuming applications
        """
        return self._failedValidationTimes

    @property
    def FailedValidationTimesUtc(self):
        return [UtcDateTime.UtcDateTime(x) for x in self._failedValidationTimes]

    @property
    def LastValidationStatus(self):
        return self._lastValidationStatus

    @LastValidationStatus.setter
    def LastValidationStatus(self, value: ValidationStatus):
        self._lastValidationStatus = value

    @property
    def FirstLaunchTime(self):
        return self._firstLaunchTime

    @FirstLaunchTime.setter
    def FirstLaunchTime(self, value: UtcDateTime):
        self._firstLaunchTime = value

    @property
    def GracePeriodForValidationFailures(self):
        return self._gracePeriodForValidationFailures

    @GracePeriodForValidationFailures.setter
    def GracePeriodForValidationFailures(self, value):
        self._gracePeriodForValidationFailures = value

    @property
    def LastSuccessfulValidationTime(self):
        if self._lastSuccessfulValidationTime is None:
            return None
        else:
            return self._lastSuccessfulValidationTime.AsDateTime()

    @LastSuccessfulValidationTime.setter
    def LastSuccessfulValidationTime(self, value):
        if isinstance(value, datetime.datetime):
            self._lastSuccessfulValidationTime = UtcDateTime.UtcDateTime(value)
        else:
            self._lastSuccessfulValidationTime = value

    @property
    def Registered(self):
        return self._registered

    @Registered.setter
    def Registered(self, value):
        self._registered = value

    @property
    def RegisteredAt(self) -> Optional[datetime.datetime]:
        if self._registeredAt is None:
            return None
        else:
            return self._registeredAt.AsDateTime()

    @property
    def RegisteredAtUtc(self):
        return self._registeredAt

    @RegisteredAt.setter
    def RegisteredAt(self, value: Union[datetime.datetime, UtcDateTime.UtcDateTime, type(None)]):
        if value is None:
            self._registeredAt = None
        elif isinstance(value, datetime.datetime):
            self._registeredAt = UtcDateTime.UtcDateTime(value)
        else:
            self._registeredAt = value

    def AdditionalUpdateAfterRegistrationSuccess(self, result):
        raise NotImplementedError()



from typing import Callable, Tuple

from .ControlStrategy import ControlStrategy
from .RegistrationResult import *
from .SessionValidity import SessionValidity
from .control_logic import ControlLogic
from .enumerations.licensing_enumerations import *
from .states.ControlState import ControlState
from .timehandling import TimeProvider
from .timehandling.UtcDateTime import UtcDateTime
from .webservice.server_results import RegisterResultBase


def InitializeState(state: ControlState):
    if state.FirstLaunchTime is None:
        state.FirstLaunchTime = GetUtcNow()


def GetUtcNow() -> UtcDateTime:
    return TimeProvider.Current().UtcNow()


def Register(state: ControlState, registration_callback: Callable[[], RegisterResultBase]):
    InitializeState(state)
    if state.Registered:
        return RegistrationResult.GetSuccess()
    result = registration_callback()
    if result.Success:
        state.Registered = True
        state.RegisteredAt = result.FirstRegisteredAt
        state.GracePeriodForValidationFailures = result.GracePeriodForValidationFailures
        state.FailedRegistrationTimes.clear()
        state.AdditionalUpdateAfterRegistrationSuccess(result)
        return RegistrationResult.GetSuccess()
    else:
        state.FailedRegistrationTimesUnsafe.insert(0, (result.ServerTime or GetUtcNow().AsDateTime()))
        return RegistrationResult.GetError(result.ErrorMessage)


def GetUnregisteredValidity(strategy: ControlStrategy, state: ControlState,
                            current_time: UtcDateTime) -> SessionValidity:
    if ControlLogic.IsFirstLaunchTimeInvalid(state.FirstLaunchTime, current_time):
        state.LastValidationStatus = ValidationStatus.UserTamperingDetected
        return SessionValidity.Invalid(InvalidReason.UserTamperingDetected)

    if ControlLogic.IsTimeInconsistencyDetected(state.FirstLaunchTime, state.FailedRegistrationTimesUtc, current_time,
                                                False):
        state.LastValidationStatus = ValidationStatus.UserTamperingDetected
        return SessionValidity.Invalid(InvalidReason.UserTamperingDetected)

    result, expiration = ControlLogic.IsInGracePeriodForUnregisteredProduct(strategy, state.FirstLaunchTime,
                                                                            current_time)
    if result:
        return SessionValidity.UnregisteredGracePeriod(expiration)
    else:
        state.LastValidationStatus = ValidationStatus.RegistrationRequired
        return SessionValidity.Invalid(InvalidReason.InstanceNotRegistered)


class OfflineGracePeriodStatus(Enum):
    UserTamperingDetected = 0,
    NotPermitted = 1,
    Available = 2


def GetServerUnreachableValidity(state: ControlState, local_time: UtcDateTime) -> SessionValidity:
    offline_status, grace_period_expiration = CheckOfflineStatus(state, local_time)
    if offline_status is OfflineGracePeriodStatus.UserTamperingDetected:
        return SessionValidity.Invalid(InvalidReason.UserTamperingDetected)
    elif offline_status is OfflineGracePeriodStatus.NotPermitted:
        return SessionValidity.Invalid(InvalidReason.ServiceUnreachable)
    else:
        return SessionValidity.ValidationFailureGracePeriod(grace_period_expiration)


def CheckOfflineStatus(state: ControlState, local_time: UtcDateTime, update_last_validation_status: bool = True) -> \
        Tuple[OfflineGracePeriodStatus, UtcDateTime]:
    if state.LastSuccessfulValidationTime is None:
        return OfflineGracePeriodStatus.NotPermitted, None
    if state.LastValidationStatus is ValidationStatus.Success:
        if IsTimeInconsistencyDetected(state, local_time):
            if update_last_validation_status:
                state.LastValidationStatus = ValidationStatus.UserTamperingDetected
            return OfflineGracePeriodStatus.UserTamperingDetected, None
        result, expiration = ControlLogic.IsInGracePeriodForValidationFailures(state.GracePeriodForValidationFailures,
                                                                               UtcDateTime(state.FailedValidationTimes[-1]),
                                                                               local_time)
        if result:
            return OfflineGracePeriodStatus.Available, expiration
        else:
            return OfflineGracePeriodStatus.NotPermitted, None
    else:
        return OfflineGracePeriodStatus.NotPermitted, None


def LogServerReached(state: ControlState, server_time: UtcDateTime, server_state: ValidationStatus):
    state.FailedValidationTimes.clear()
    state.LastSuccessfulValidationTime = server_time
    state.LastValidationStatus = server_state


def LogServerNotReached(state: ControlState, local_time: UtcDateTime):
    state.FailedValidationTimesUnsafe.insert(0, local_time.AsDateTime())


def IsTimeInconsistencyDetected(state: ControlState, local_time: UtcDateTime) -> bool:
    return ControlLogic.IsTimeInconsistencyDetected(state.LastSuccessfulValidationTime, state.FailedValidationTimesUtc,
                                                    local_time, False)


def StatusToReason(status: ValidationStatus) -> InvalidReason:
    if status is ValidationStatus.RegistrationRequired:
        return InvalidReason.InstanceNotRegistered
    try:
        return InvalidReason[status.name]
    except KeyError:
        return InvalidReason.Unknown



from datetime import *
from typing import List, Optional
from ..enumerations.licensing_enumerations import ValidationStatus
from uuid import UUID, uuid4


class ServerResult(object):
    def __init__(self):
        self._errorMessage = ''

    @property
    def ErrorMessage(self):
        return self._errorMessage

    @ErrorMessage.setter
    def ErrorMessage(self, value: str):
        self._errorMessage = value


class TimedServerResult(ServerResult):
    def __init__(self):
        super().__init__()
        self._serverTime = None

    @property
    def ServerTime(self):
        return self._serverTime

    @ServerTime.setter
    def ServerTime(self, value: datetime):
        self._serverTime = value


class RegisterResultBase(TimedServerResult):
    def __init__(self):
        super().__init__()
        self._firstRegisteredAt = None
        self._gracePeriodForValidationFailures = timedelta(hours=0)
        self._success = False

    @property
    def Success(self):
        """ Gets whether registration was successful. """
        return self._success

    @Success.setter
    def Success(self, value: bool):
        self._success = value

    @property
    def GracePeriodForValidationFailures(self):
        """Gets the grace period allowed for validation failures."""
        return self._gracePeriodForValidationFailures

    @GracePeriodForValidationFailures.setter
    def GracePeriodForValidationFailures(self, value: timedelta):
        self._gracePeriodForValidationFailures = value

    @property
    def FirstRegisteredAt(self):
        """
        Gets the time at which this instance and/or user was *first* registered.  Multiple registrations
        of an instance are valid and will succeed, but this property will always return
        the time of the very first registration.  This is useful for determining the length
        of trial periods, for example.
        """
        return self._firstRegisteredAt

    @FirstRegisteredAt.setter
    def FirstRegisteredAt(self, value: Optional[datetime]):
        self._firstRegisteredAt = value


class RegisterInstanceResult(RegisterResultBase):

    def __init__(self):
        super().__init__()
        self._InstanceId = None

    @property
    def InstanceId(self):
        return self._InstanceId

    @InstanceId.setter
    def InstanceId(self, value: UUID):
        self._InstanceId = value


class RegisterResultWithSessionDuration(RegisterResultBase):

    def __init__(self):
        super().__init__()
        self._SessionDuration = timedelta(hours=0)

    @property
    def SessionDuration(self):
        return self._SessionDuration

    @SessionDuration.setter
    def SessionDuration(self, value: timedelta):
        self._SessionDuration = value


class RegisterUserResult(RegisterResultWithSessionDuration):

    def __init__(self):
        super().__init__()
        self._UserId = None

    @property
    def UserId(self):
        return self._UserId

    @UserId.setter
    def UserId(self, value: UUID):
        self._UserId = value


class RegisterAppResult(RegisterUserResult):

    def __init__(self):
        super().__init__()
        self._InstanceId = None

    @property
    def InstanceId(self):
        return self._InstanceId

    @InstanceId.setter
    def InstanceId(self, value: UUID):
        self._InstanceId = value

class GetUserLicenseServerResult(TimedServerResult):
    def __init__(self):
        super().__init__()
        self._LicenseKeys = []
    
    @property
    def LicenseKeys(self):
        return self._LicenseKeys

    @LicenseKeys.setter
    def LicenseKeys(self, value: List[str]):
        self._LicenseKeys = value

class ValidatedResult(TimedServerResult):

    def __init__(self):
        super().__init__()
        self._Status = ValidationStatus.RegistrationRequired

    @property
    def Status(self):
        return self._Status

    @Status.setter
    def Status(self, value: ValidationStatus):
        self._Status = value


class ValidatedSessionResult(ValidatedResult):

    def __init__(self):
        super().__init__()
        self._SessionDuration = timedelta(hours=0)
        self._GracePeriodForValidationFailures = timedelta(days=0)
        self._Features = []
        self._AuthToken = ''

    @property
    def AuthToken(self):
        return self._AuthToken

    @AuthToken.setter
    def AuthToken(self, value: str):
        self._AuthToken = value

    @property
    def Features(self):
        return self._Features

    @Features.setter
    def Features(self, value: List[str]):
        self._Features = value

    @property
    def GracePeriodForValidationFailures(self):
        return self._GracePeriodForValidationFailures

    @GracePeriodForValidationFailures.setter
    def GracePeriodForValidationFailures(self, value: timedelta):
        self._GracePeriodForValidationFailures = value

    @property
    def SessionDuration(self):
        return self._SessionDuration

    @SessionDuration.setter
    def SessionDuration(self, value: timedelta):
        self._SessionDuration = value


class ValidateInstanceResult(ValidatedResult):

    def __init__(self):
        super().__init__()
        self._Features = []
        self._GracePeriodForValidationFailures = []

    @property
    def GracePeriodForValidationFailures(self):
        return self._GracePeriodForValidationFailures

    @GracePeriodForValidationFailures.setter
    def GracePeriodForValidationFailures(self, value: timedelta):
        self._GracePeriodForValidationFailures = value

    @property
    def Features(self):
        return self._Features

    @Features.setter
    def Features(self, value):
        self._Features = value


class BeginSessionResult(ValidatedSessionResult):

    def __init__(self):
        super().__init__()
        self._SessionKey = ''

    @property
    def SessionKey(self):
        return self._SessionKey

    @SessionKey.setter
    def SessionKey(self, value):
        self._SessionKey = value


class BeginAppSessionResult(BeginSessionResult):

    def __init__(self):
        super().__init__()
        self._Components = []
        self._ComponentEntitlements = []

    @property
    def Components(self):
        return self._Components

    @Components.setter
    def Components(self, value):
        self._Components = value

    @property
    def ComponentEntitlements(self):
        return self._ComponentEntitlements

    @ComponentEntitlements.setter
    def ComponentEntitlements(self, value):
        self._ComponentEntitlements = value


class CheckOutComponentResult(ValidatedResult):

    def __init__(self):
        super().__init__()
        self._Components = []
        self._ComponentEntitlements = []

    @property
    def Components(self):
        return self._Components

    @Components.setter
    def Components(self, value):
        self._Components = value

    @property
    def ComponentEntitlements(self):
        return self._ComponentEntitlements

    @ComponentEntitlements.setter
    def ComponentEntitlements(self, value):
        self._ComponentEntitlements = value


class EndSessionResult(TimedServerResult):

    def __init__(self):
        super().__init__()
        self._Success = False

    @property
    def Success(self):
        return self._Success

    @Success.setter
    def Success(self, value):
        self._Success = value


class ComponentsStatusResult(ServerResult):

    def __init__(self):
        super().__init__()
        self._ComponentEntitlements = []
        self._Components = []
        self._Success = False

    @property
    def Success(self):
        return self._Success

    @Success.setter
    def Success(self, value):
        self._Success = value

    @property
    def Components(self):
        return self._Components

    @Components.setter
    def Components(self, value):
        self._Components = value

    @property
    def ComponentEntitlements(self):
        return self._ComponentEntitlements

    @ComponentEntitlements.setter
    def ComponentEntitlements(self, value):
        self._ComponentEntitlements = value



import json
from collections import namedtuple
from typing import Dict, Optional

import requests
from Crypto.Cipher import PKCS1_v1_5
from Crypto.PublicKey import RSA

from . import WebHelper, ConnectionSettings
from .jsoncomponent import JSONComponent
from .jsoncomponententitlement import JSONComponentEntitlement
from .server_results import *
from ..UserIdentity import UserIdentity
from .. import InstanceIdentity
from ..InstanceIdentity import InstanceIdentity as II
from ..UserInfo import UserInfo
from ..enumerations.licensing_enumerations import ValidationStatus


class LicensingWebService(object):
    def __init__(self, public_key: str):
        self._publicKey = public_key

    @staticmethod
    def AddCommonData(post_data: Dict[str, object], product_id: UUID, instance_identity: II,
                      user_identity: UserIdentity, user_info: Optional[UserInfo] = None):
        post_data['product_guid'] = str(product_id)
        if not (user_identity is None):
            post_data['user_identity'] = user_identity.Data

        if not (instance_identity is None):
            instance_identity_data = dict()
            if instance_identity.Version != InstanceIdentity.DEFAULT_VERSION:
                instance_identity_data['_version'] = instance_identity.Version
            if instance_identity.MaximumMismatchCount != InstanceIdentity.DEFAULT_MAX_MISMATCH_COUNT:
                instance_identity_data['_max_mismatch_count'] = instance_identity.MaximumMismatchCount

            for key, value in instance_identity.Values.items():
                key_data = dict()
                key_data['value'] = value.Value
                if value.Weight != InstanceIdentity.DEFAULT_WEIGHT:
                    key_data['_weight'] = value.Weight
                if value.CaseSensitive != InstanceIdentity.DEFAULT_CASE_SENSITIVE:
                    key_data['_case_sensitive'] = value.CaseSensitive
                if value.MaxArrayValueDifference != InstanceIdentity.DEFAULT_MAX_ARRAY_VALUE_DIFFERENCE:
                    key_data['_max_array_value_difference'] = value.MaxArrayValueDifference
                instance_identity_data[key] = key_data
            post_data['instance_identity'] = instance_identity_data

        if not (user_info is None):
            user_info_data = dict()
            user_info_data['name'] = user_info.Name
            user_info_data['email'] = user_info.Email
            post_data['user_info'] = user_info_data

    @staticmethod
    def ParseResultDateTime(json_result: any, field: str):
        return datetime.strptime(json_result[field], "%a, %d %b %Y %H:%M:%S %z")

    @staticmethod
    def GetUri(sub_uri: str) -> str:
        host = ConnectionSettings.ConnectionSettings.Host()
        if host.endswith("/"):
            host = host.rstrip("/")
        return host + "/api/v3.2/auth/" + sub_uri

    def GetRSAProvider(self):
        rsa_key = RSA.import_key(self._publicKey)
        # noinspection PyTypeChecker
        return PKCS1_v1_5.new(rsa_key)

    def RegisterAppAsync(self, product_id: UUID, license_key: str, user_info: UserInfo,
                         instance_identity: II, user_identity: UserIdentity) -> RegisterAppResult:
        post_data = dict()
        LicensingWebService.AddCommonData(post_data, product_id, instance_identity, user_identity, user_info)
        post_data['license_key'] = license_key
        uri = LicensingWebService.GetUri("register_app.json")
        rsa_provider = self.GetRSAProvider()
        try:
            result = WebHelper.WebHelper.PostEncryptedAsync(requests, uri, product_id, post_data, rsa_provider, ssl_verify=ConnectionSettings.ConnectionSettings.SSLVerify())
            json_result = json.JSONDecoder().decode(result)
            if json_result["success"]:
                app_result = RegisterAppResult()
                app_result.Success = True
                app_result.ServerTime = LicensingWebService.ParseResultDateTime(json_result, "time")
                app_result.InstanceId = UUID(hex=json_result["instance_guid"])
                app_result.GracePeriodForValidationFailures = timedelta(seconds=json_result["validation_grace_period"])
                app_result.SessionDuration = timedelta(seconds=json_result["session_duration"])
                app_result.FirstRegisteredAt = LicensingWebService.ParseResultDateTime(json_result,
                                                                                       "first_registered_at")
                app_result.UserId = UUID(hex=json_result["user_guid"])
                return app_result
            else:
                app_result = RegisterAppResult()
                app_result.Success = False
                app_result.ServerTime = LicensingWebService.ParseResultDateTime(json_result, "time")
                app_result.ErrorMessage = json_result["error"]
                return app_result
        except ConnectionError as conn_error:
            app_result = RegisterAppResult()
            app_result.Success = False
            app_result.ErrorMessage = "Error contacting registration service. " + conn_error.strerror
            return app_result
        except Exception as gen_error:
            app_result = RegisterAppResult()
            app_result.Success = False
            app_result.ErrorMessage = "A general error occurred during registration. \
                Please contact the software vendor. " + str(gen_error)
            return app_result

    def BeginAppSessionAsync(self, product_id: UUID, instance_id: UUID, version: str, user_id: Optional[UUID],
                             instance_identity: II = None, user_identity: UserIdentity = None,
                             requested_session_duration: Optional[timedelta] = None) -> BeginAppSessionResult:
        post_data = dict()
        LicensingWebService.AddCommonData(post_data, product_id, instance_identity, user_identity)
        post_data["instance_guid"] = str(instance_id)
        post_data["version"] = version
        if not (user_id is None):
            post_data["user_guid"] = str(user_id)
        if not (requested_session_duration is None):
            post_data["duration"] = str(requested_session_duration.seconds)
        uri = LicensingWebService.GetUri("begin_app_session.json")
        rsa_provider = self.GetRSAProvider()
        try:
            result = WebHelper.WebHelper.PostEncryptedAsync(requests, uri, product_id, post_data, rsa_provider, ssl_verify=ConnectionSettings.ConnectionSettings.SSLVerify())
            json_result = json.JSONDecoder().decode(result)
            status = ValidationStatus(json_result["status"])
            if status is ValidationStatus.Success:
                app_result = BeginAppSessionResult()
                app_result.Status = status
                app_result.ServerTime = LicensingWebService.ParseResultDateTime(json_result, "time")
                app_result.SessionKey = json_result["session_key"]
                app_result.Features = json_result["features"] or []
                # Deserialize component info if available
                LicensingWebService.parse_components(app_result, json_result)
                # Deserialize component entitlement info if available
                LicensingWebService.parse_component_entitlements(app_result, json_result)
                app_result.AuthToken = json_result["token"]
                app_result.GracePeriodForValidationFailures = timedelta(seconds=json_result["validation_grace_period"])
                app_result.SessionDuration = timedelta(seconds=json_result["session_duration"])
                return app_result
            else:
                app_result = BeginAppSessionResult()
                app_result.Status = status
                app_result.ServerTime = LicensingWebService.ParseResultDateTime(json_result, "time")
                app_result.ErrorMessage = json_result["error"]
                return app_result
        except ConnectionError as conn_error:
            app_result = BeginAppSessionResult()
            app_result.Status = ValidationStatus.ServiceUnreachable
            app_result.ErrorMessage = conn_error.strerror
            return app_result
        except Exception as gen_error:
            app_result = BeginAppSessionResult()
            app_result.Status = ValidationStatus.GeneralError
            app_result.ErrorMessage = str(gen_error)
            return app_result
    def GetUserLicense(self, product_id: UUID, user_identity: UserIdentity) -> GetUserLicenseServerResult:
        post_data = dict()
        LicensingWebService.AddCommonData(post_data, product_id, None, user_identity, None)
        uri = LicensingWebService.GetUri("get_user_license.json")
        rsa_provider = self.GetRSAProvider()
        try:
            result = WebHelper.WebHelper.PostEncryptedAsync(requests, uri, product_id, post_data, rsa_provider, ssl_verify=ConnectionSettings.ConnectionSettings.SSLVerify())
            json_result = json.JSONDecoder().decode(result)
            if json_result["success"]:
                app_result = GetUserLicenseServerResult()
                app_result.Success = True
                app_result.ServerTime = LicensingWebService.ParseResultDateTime(json_result, "time")
                app_result.Licenses = json_result["licenses"]
                return app_result
            else:
                app_result = GetUserLicenseServerResult()
                app_result.Success = False
                app_result.ServerTime = LicensingWebService.ParseResultDateTime(json_result, "time")
                app_result.ErrorMessage = json_result["error"]
                app_result.Licenses = []
                return app_result
        except ConnectionError as conn_error:
            app_result = GetUserLicenseServerResult()
            app_result.Success = False
            app_result.ErrorMessage = "Error contacting registration service. " + conn_error.strerror
            return app_result
        except Exception as gen_error:
            app_result = GetUserLicenseServerResult()
            app_result.Success = False
            app_result.ErrorMessage = "A general error occurred during registration. \
                Please contact the software vendor. " + str(gen_error)
            return app_result
    @staticmethod
    def parse_component_entitlements(app_result, json_result):
        if json_result.keys().__contains__("component_entitlements"):
            json_component_entitlements: List[JSONComponentEntitlement] = [
                JSONComponentEntitlement(c) for c in json_result["component_entitlements"]]
            app_result.ComponentEntitlements = [jc.Deserialize() for jc in json_component_entitlements]
        else:
            app_result.ComponentEntitlements = []

    @staticmethod
    def parse_components(app_result, json_result):
        if json_result.keys().__contains__("components"):
            json_components: List[JSONComponent] = [JSONComponent(c) for c in json_result["components"]]
            app_result.Components = [jc.Deserialize() for jc in json_components]
        else:
            app_result.Components = []

    def ExtendSession(self, product_id: UUID, token: str,
                      requested_extension_duration: Optional[timedelta] = None) -> ValidatedSessionResult:
        post_data = dict()
        LicensingWebService.AddCommonData(post_data, product_id, None, None, None)
        post_data["token"] = token
        if requested_extension_duration is not None:
            post_data["duration"] = str(requested_extension_duration.seconds)
        uri = LicensingWebService.GetUri("extend_session.json")
        rsa_provider = self.GetRSAProvider()
        try:
            result = WebHelper.WebHelper.PostEncryptedAsync(requests, uri, product_id, post_data, rsa_provider, ssl_verify=ConnectionSettings.ConnectionSettings.SSLVerify())
            json_result = json.JSONDecoder().decode(result)
            status = ValidationStatus(json_result["status"])
            if status is ValidationStatus.Success:
                app_result = ValidatedSessionResult()
                app_result.Status = status
                app_result.ServerTime = LicensingWebService.ParseResultDateTime(json_result, "time")
                app_result.Features = json_result["features"] or []
                app_result.AuthToken = json_result["token"]
                app_result.GracePeriodForValidationFailures = timedelta(seconds=json_result["validation_grace_period"])
                app_result.SessionDuration = timedelta(seconds=json_result["session_duration"])
                return app_result
            else:
                app_result = ValidatedSessionResult()
                app_result.Status = status
                app_result.ServerTime = LicensingWebService.ParseResultDateTime(json_result, "time")
                app_result.ErrorMessage = json_result["error"]
                return app_result
        except ConnectionError as conn_error:
            app_result = ValidatedSessionResult()
            app_result.Status = ValidationStatus.ServiceUnreachable
            app_result.ErrorMessage = conn_error.strerror
            return app_result
        except Exception as gen_error:
            app_result = ValidatedSessionResult()
            app_result.Status = ValidationStatus.GeneralError
            app_result.ErrorMessage = str(gen_error)
            return app_result

    def CheckOutComponentCommon(self, post_data: dict, uri_part: str, product_id: str) -> CheckOutComponentResult:
        uri = LicensingWebService.GetUri(uri_part)
        rsa_provider = self.GetRSAProvider()
        try:
            result = WebHelper.WebHelper.PostEncryptedAsync(requests, uri, product_id, post_data, rsa_provider, ssl_verify=ConnectionSettings.ConnectionSettings.SSLVerify())
            json_result = json.JSONDecoder().decode(result)
            status = ValidationStatus(json_result["status"])
            if status is ValidationStatus.Success:
                app_result = CheckOutComponentResult()
                app_result.Status = status
                app_result.ServerTime = LicensingWebService.ParseResultDateTime(json_result, "time")
                LicensingWebService.parse_components(app_result, json_result)
                LicensingWebService.parse_component_entitlements(app_result, json_result)
                return app_result
            else:
                app_result = CheckOutComponentResult()
                app_result.Status = status
                app_result.ServerTime = LicensingWebService.ParseResultDateTime(json_result, "time")
                app_result.ErrorMessage = json_result["error"]
                return app_result
        except ConnectionError as conn_error:
            app_result = CheckOutComponentResult()
            app_result.Status = ValidationStatus.ServiceUnreachable
            app_result.ErrorMessage = conn_error.strerror
            return app_result
        except Exception as gen_error:
            app_result = CheckOutComponentResult()
            app_result.Status = ValidationStatus.GeneralError
            app_result.ErrorMessage = str(gen_error)
            return app_result

    def CheckOutComponents(self, product_id: UUID, session_key: str, components: List[str]) -> CheckOutComponentResult:
        post_data = dict()
        LicensingWebService.AddCommonData(post_data, product_id, None, None, None)
        post_data["session_key"] = session_key
        post_data["components"] = components 
        return self.CheckOutComponentCommon(post_data, "check_out_components.json", product_id)
        
    def CheckOutConsumableComponent(self, product_id: UUID, session_key: str, component: str, token_count: Optional[int] = None) -> CheckOutComponentResult:
        post_data = dict()
        LicensingWebService.AddCommonData(post_data, product_id, None, None, None)
        post_data["session_key"] = session_key
        post_data["component"] = component
        post_data["token_count"] = token_count
        return self.CheckOutComponentCommon(post_data, "check_out_consumable_component.json", product_id)

    def EndSession(self, product_id: UUID, token: str) -> EndSessionResult:
        post_data = dict()
        LicensingWebService.AddCommonData(post_data, product_id, None, None, None)
        post_data["token"] = token
        uri = LicensingWebService.GetUri("end_session.json")
        rsa_provider = self.GetRSAProvider()
        try:
            result = WebHelper.WebHelper.PostEncryptedAsync(requests, uri, product_id, post_data, rsa_provider, ssl_verify=ConnectionSettings.ConnectionSettings.SSLVerify())
            json_result = json.JSONDecoder().decode(result)
            app_result = EndSessionResult()
            app_result.Success = True
            app_result.ServerTime = LicensingWebService.ParseResultDateTime(json_result, "time")
            return app_result
        except ConnectionError as conn_error:
            app_result = EndSessionResult()
            app_result.Success = False
            app_result.ErrorMessage = conn_error.strerror
            return app_result
        except Exception as gen_error:
            app_result = EndSessionResult()
            app_result.Success = False
            app_result.ErrorMessage = str(gen_error)
            return app_result

    def GetComponentStatus(self, product_id: UUID, session_key: str) -> ComponentsStatusResult:
        post_data = dict()
        LicensingWebService.AddCommonData(post_data, product_id, None, None, None)
        post_data["session_key"] = session_key
        uri = LicensingWebService.GetUri("components_status.json")
        rsa_provider = self.GetRSAProvider()
        try:
            result = WebHelper.WebHelper.PostEncryptedAsync(requests, uri, product_id, post_data, rsa_provider, ssl_verify=ConnectionSettings.ConnectionSettings.SSLVerify())
            json_result = json.JSONDecoder().decode(result)
            app_result = ComponentsStatusResult()
            app_result.Success = True
            LicensingWebService.parse_components(app_result, json_result)
            LicensingWebService.parse_component_entitlements(app_result, json_result)
            return app_result
        except ConnectionError as conn_error:
            app_result = ComponentsStatusResult()
            app_result.Success = False
            app_result.ErrorMessage = conn_error.strerror
            return app_result
        except Exception as gen_error:
            app_result = ComponentsStatusResult()
            app_result.Success = False
            app_result.ErrorMessage = str(gen_error)
            return app_result



from ..enumerations.licensing_enumerations import SessionState
from uuid import UUID


class ValidatedSessionState:
    _authToken: str
    _sessionKey: str
    _sessionState: SessionState
    # Not supported yet
    _offlineCheckoutKey: str
    _offlineSessionID: UUID

    def __init__(self):
        self._authToken = ""
        self._sessionKey = ""
        self._sessionState = SessionState.NotStarted
        self._offlineCheckoutKey = ""
        self._offlineSessionID = None

    @property
    def AuthToken(self) -> str:
        return self._authToken

    @AuthToken.setter
    def AuthToken(self, value: str):
        setattr(self, "_authToken", value)

    @property
    def SessionKey(self) -> str:
        return self._sessionKey

    @SessionKey.setter
    def SessionKey(self, value: str):
        setattr(self, "_sessionKey", value)

    @property
    def SessionState(self) -> SessionState:
        return self._sessionState

    @SessionState.setter
    def SessionState(self, value: SessionState):
        setattr(self, "_sessionState", value)

    @property
    def OfflineCheckoutKey(self) -> str:
        return self._offlineCheckoutKey

    @OfflineCheckoutKey.setter
    def OfflineCheckoutKey(self, value: str):
        setattr(self, "_offlineCheckoutKey", value)

    @property
    def OfflineSessionID(self) -> UUID:
        return self._offlineSessionID

    @OfflineSessionID.setter
    def OfflineSessionID(self, value: UUID):
        setattr(self, "_offlineSessionID", value)


def fromOther(source: ValidatedSessionState) -> ValidatedSessionState:
    dest = ValidatedSessionState()
    if source is None:
        dest.SessionState = SessionState.NotStarted
        return dest
    dest.AuthToken = source.AuthToken
    dest.SessionKey = source.SessionKey
    dest.SessionState = source.SessionState
    dest.OfflineCheckoutKey = source.OfflineCheckoutKey
    dest.OfflineSessionID = source.OfflineSessionID
    return dest



from Crypto.PublicKey import RSA
from Crypto.Cipher import AES, PKCS1_v1_5
from Crypto.Random import get_random_bytes
from Crypto.Util import Padding
import json
import base64
import requests
from typing import Optional, Dict
from uuid import UUID


class ServerReachedWithNonSuccessErrorCodeError(IOError):
    def __init__(self, message: str):
        self._message = message

    @property
    def Message(self):
        return self._message


class WebHelper(object):
    @staticmethod
    def PostEncryptedAsync(client: requests, uri: str, product_id: UUID, data: Dict[str, object],
                           rsa_provider: any, ssl_verify:bool = True):
        key, iv, encrypted_content = WebHelper.CreateEncryptedJsonContent(product_id, data, rsa_provider)
        headers = {'cache-control': "no-cache", 'content-type': "application/json"}
        postResult = client.request("POST", uri, json=encrypted_content, headers=headers, verify=True)
        if postResult.status_code != 200:
            raise ServerReachedWithNonSuccessErrorCodeError("Unsuccessful call to server");
        crypted_response = base64.decodebytes(postResult.content)
        return WebHelper.DecryptResponse(crypted_response, key, iv)

    @staticmethod
    def CreateEncryptedJsonContent(product_id: UUID, data: Dict[str, object], rsa_provider: any):
        payload_data = dict()
        payload_data["key"] = str(product_id)
        key, iv, data_string = WebHelper.EncryptData(data, rsa_provider)
        payload_data["data"] = data_string
        return key, iv, payload_data

    @staticmethod
    def EncryptData(data: Dict[str, object], rsa_provider: any):
        key = get_random_bytes(16)
        iv = get_random_bytes(16)
        cipher_aes = AES.new(key=key, iv=iv, mode=AES.MODE_CBC)
        encrypted_aes_key = rsa_provider.encrypt(key)
        json_encoder = json.JSONEncoder()
        post_string = json_encoder.encode(data)
        input_buffer = post_string.encode("utf-8")
        input_buffer = Padding.pad(input_buffer, AES.block_size)
        crypted_data = cipher_aes.encrypt(input_buffer)
        padding_char = b"\x00"
        final_buffer = b"".join([encrypted_aes_key, padding_char, iv, padding_char, crypted_data])
        return key, cipher_aes.IV, base64.b64encode(final_buffer).decode('utf-8')

    @staticmethod
    def DecryptResponse(crypted_response: bytes, key: bytes, iv: bytes):
        cipher_aes = AES.new(key=key, iv=iv, mode=AES.MODE_CBC)
        padded_response = cipher_aes.decrypt(crypted_response)
        final_json_string = Padding.unpad(padded_response, AES.block_size).decode('utf-8')
        return final_json_string



from .SessionControlState import SessionControlState
from uuid import UUID
from ..webservice import server_results
from ..desktop.OfflineSessionInfo import OfflineSessionInfo


class ValidationState(SessionControlState):
    def __init__(self):
        super(ValidationState, self).__init__()
        self._components = []
        self._componentEntitlements = []
        self._componentEntitlementsLoaded = False
        self._componentsLoaded = False
        self._userId = None
        self._instanceId = None
        self._offlineSessions = []

    @staticmethod
    def fromOther(source, dest):
        if dest is None:
            dest = ValidationState()
        if source is None:
            return dest
        if not isinstance(source, ValidationState):
            raise ValueError("source must also be a ValidationState object")
        dest = SessionControlState.fromOther(source, dest)
        dest._components = source.Components
        dest._componentEntitlements = source.ComponentEntitlements
        dest._offlineSessions = list(source._offlineSessions)

        if dest.Registered:
            dest.InstanceId = source.InstanceId
            dest.UserId = source.UserId
            dest.ComponentsLoaded = source.ComponentsLoaded
            dest.ComponentEntitlementsLoaded = source.ComponentEntitlementsLoaded
        return dest

    @property
    def InstanceId(self) -> UUID:
        """
        Gets the instance id of the registered instance
        :rtype: UUID
        """
        return self._instanceId

    @InstanceId.setter
    def InstanceId(self, value: UUID):
        setattr(self, "_instanceId", value)

    @property
    def UserId(self):
        return self._userId

    @UserId.setter
    def UserId(self, value):
        setattr(self, "_userId", value)

    @property
    def ComponentsLoaded(self):
        return self._componentsLoaded

    @ComponentsLoaded.setter
    def ComponentsLoaded(self, value):
        setattr(self, "_componentsLoaded", value)

    @property
    def ComponentEntitlementsLoaded(self):
        return self._componentEntitlementsLoaded

    @ComponentEntitlementsLoaded.setter
    def ComponentEntitlementsLoaded(self, value):
        setattr(self, "_componentEntitlementsLoaded", value)

    Components = property(lambda self: self._components)
    ComponentEntitlements = property(lambda self: self._componentEntitlements)
    ComponentsRO = property(lambda self: list(self._components))
    ComponentEntitlementsRO = property(lambda self: list(self._componentEntitlements))
    ComponentsIfLoaded = property(lambda self: list(self._components) if self.ComponentsLoaded else [])
    ComponentEntitlementsIfLoaded = property(lambda self:
                                             list(self._componentEntitlements)
                                             if self.ComponentEntitlementsLoaded else [])
    OfflineSessions = property(lambda self: list(self._offlineSessions))
    OfflineSessionsUnsafe = property(lambda self: self._offlineSessions)
    def AdditionalUpdateAfterRegistrationSuccess(self, result: server_results.RegisterAppResult):
        super().AdditionalUpdateAfterRegistrationSuccess(result)
        self._instanceId = result.InstanceId
        self._userId = result.UserId
        self.SessionDuration = result.SessionDuration

    def OfflineSession(self, guid: UUID) -> OfflineSessionInfo:
        return next((x for x in self._offlineSessions if x.LocalSessionId == guid), None)



from datetime import datetime
from .enumerations.licensing_enumerations import ComponentLicenseModel
from .timehandling.UtcDateTime import UtcDateTime


class ComponentInfo:
    def __init__(self):
        self._name = None
        self._licenseModel = ComponentLicenseModel.Nothing
        self._tokensRequired = 0
        self._hasFreeTrial = False
        self._freeTrialExpirationTime = datetime.min
        self._originalTokensRequired = 0

    @property
    def Name(self) -> str:
        return self._name

    @Name.setter
    def Name(self, value: str):
        self._name = value

    @property
    def LicenseModel(self) -> ComponentLicenseModel:
        return self._licenseModel

    @LicenseModel.setter
    def LicenseModel(self, value: ComponentLicenseModel):
        self._licenseModel = value

    @property
    def TokensRequired(self) -> int:
        return self._tokensRequired

    @TokensRequired.setter
    def TokensRequired(self, value: int):
        self._tokensRequired = value

    @property
    def HasFreeTrial(self) -> bool:
        return self._hasFreeTrial

    @HasFreeTrial.setter
    def HasFreeTrial(self, value: bool):
        self._hasFreeTrial = value

    @property
    def FreeTrialExpirationTime(self) -> datetime:
        return self._freeTrialExpirationTime

    @FreeTrialExpirationTime.setter
    def FreeTrialExpirationTime(self, value: datetime):
        self._freeTrialExpirationTime = value

    @property
    def OriginalTokensRequired(self) -> int:
        return self._originalTokensRequired

    @OriginalTokensRequired.setter
    def OriginalTokensRequired(self, value: int):
        self._originalTokensRequired = value

    def TokensRequiredAt(self, time: UtcDateTime):
        if self.HasFreeTrial:
            if self.FreeTrialExpirationTime > time.AsDateTime():
                return self.TokensRequired
            else:
                return self.OriginalTokensRequired
        else:
            return self.TokensRequired



from datetime import timedelta


class ControlStrategy:
    _locked: bool
    _gracePeriod: timedelta = timedelta(seconds=0)

    def Lock(self):
        self._locked = True

    @property
    def GracePeriodForUnregisteredProduct(self) -> timedelta:
        return self._gracePeriod

    @GracePeriodForUnregisteredProduct.setter
    def GracePeriodForUnregisteredProduct(self, value: timedelta):
        if self._locked:
            raise ValueError('Control strategy cannot be changed after it has been used')
        self._gracePeriod = value



class UserInfo(object):
    def __init__(self, name: str, email: str):
        self._name = name
        self._email = email

    @property
    def Name(self):
        return self._name

    @Name.setter
    def Name(self, value: str):
        self._name = value

    @property
    def Email(self):
        return self._email

    @Email.setter
    def Email(self, value: str):
        self._email = value

    def Verify(self):
        if self._name is None:
            raise ValueError('User name cannot be None')
        if self._email is None:
            raise ValueError('User email cannot be None')



from datetime import datetime, timedelta
import pytz


# noinspection PyProtectedMember
class UtcDateTime(object):
    def __init__(self, datetime_value: datetime):
        self._value = datetime_value.astimezone(pytz.utc)

    @staticmethod
    def Coerce(other):
        if (other is None):
            return None
        elif (isinstance(other, datetime)):
            return UtcDateTime(other)
        elif (isinstance(other, UtcDateTime)):
            return other
        else:
            return None
    def __add__(self, delta: timedelta):
        return UtcDateTime(self._value + delta)

    def __sub__(self, delta_or_other):
        if isinstance(delta_or_other, timedelta):
            return UtcDateTime(self._value - delta_or_other)
        elif isinstance(delta_or_other, UtcDateTime):
            return self._value - delta_or_other._value
        else:
            raise ValueError("delta_or_other must be a timedelta object or UtcDateTime object")

    def __lt__(self, other):
        other = UtcDateTime.Coerce(other)
        return self._value < other._value

    def __gt__(self, other):
        other = UtcDateTime.Coerce(other)
        return self._value > other._value

    def __le__(self, other):
        other = UtcDateTime.Coerce(other)
        return self._value <= other._value

    def __ge__(self, other):
        other = UtcDateTime.Coerce(other)
        return self._value >= other._value

    def __eq__(self, other):
        other = UtcDateTime.Coerce(other)
        return self._value == other._value

    def __ne__(self, other):
        other = UtcDateTime.Coerce(other)
        return self._value != other._value

    def AddDays(self, value):
        return UtcDateTime(self._value + timedelta(days=value))

    def AddMinutes(self, value):
        return UtcDateTime(self._value + timedelta(minutes=value))

    def AsDateTime(self):
        return self._value

    def __hash__(self):
        return hash(self._value)



import setuptools
import sys
# noinspection SpellCheckingInspection
with open("README.md", "r") as fh:
    long_description = fh.read()
major, minor, micro = sys.version_info[:3]
ir = ['pycryptodome', 'pytz']
if minor == 6:
    ir.append('dataclasses')

setuptools.setup(
    name="evoleap_licensing",
    version="1.1.2",
    author="evoleap",
    author_email="info@evoleap.com",
    packages=setuptools.find_packages(),
    url="https://github.com/evoleap/evoleap.licensing-python",
    license='https://evoleap.com/elm_client_api_net_eula.html',
    description='Software licensing client API for elm, a cloud-based, enterprise grade software licensing platform.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    zip_safe=True,
    install_requires=ir,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent"
    ]
)



from .enumerations.licensing_enumerations import ComponentCheckoutFailureReason


class ComponentCheckoutResult:
    _success: bool
    _failureReason: ComponentCheckoutFailureReason
    _processedOffline: bool

    def __init__(self, success: bool, failure_reason: ComponentCheckoutFailureReason, processed_offline: bool):
        self._success = success
        self._failureReason = failure_reason
        self._processedOffline = processed_offline

    @property
    def Success(self) -> bool:
        return self._success

    @property
    def FailureReason(self) -> ComponentCheckoutFailureReason:
        return self._failureReason

    @property
    def ProcessedOffline(self) -> bool:
        return self._processedOffline

    @classmethod
    def GetSuccess(cls, processed_offline: bool):
        return cls(True, ComponentCheckoutFailureReason.NoFailure, processed_offline)

    @classmethod
    def GetFailure(cls, reason: ComponentCheckoutFailureReason, processed_offline: bool):
        return cls(False, reason, processed_offline)



from datetime import timedelta
from typing import Optional, List
from uuid import UUID

from .AppSessionManager import AppSessionManager
from .. import ControlManagerHelper
from ..ComponentCheckoutResult import ComponentCheckoutResult
from ..ControlStrategy import ControlStrategy
from ..InstanceIdentity import InstanceIdentity
from ..RegistrationResult import RegistrationResult
from ..SessionValidity import SessionValidity
from ..UserIdentity import UserIdentity
from ..UserInfo import UserInfo
from ..UserLicenseResult import UserLicenseResult
from ..enumerations.licensing_enumerations import SessionState
from ..states.ValidatedSessionState import ValidatedSessionState, fromOther as fromOtherValidatedSessionState
from ..states.ValidationState import ValidationState
from ..timehandling import TimeProvider
from ..webservice.LicensingWebService import LicensingWebService
from ..webservice.server_results import RegisterResultBase


class ControlManager:
    _state: ValidationState
    _sessionState: ValidatedSessionState
    _productId: UUID
    _version: str
    _sessionManager: AppSessionManager
    _webService: LicensingWebService
    _userIdentity: UserIdentity
    _instanceIdentity: InstanceIdentity
    _strategy: ControlStrategy

    @property
    def State(self) -> ValidationState:
        return self._state

    @property
    def SessionState(self) -> ValidatedSessionState:
        return self._sessionState

    @property
    def ProductId(self) -> UUID:
        return self._productId

    @property
    def Version(self) -> str:
        return self._version

    @property
    def SessionManager(self) -> AppSessionManager:
        return self._sessionManager

    @property
    def WebService(self) -> LicensingWebService:
        return self._webService

    @property
    def UserIdentity(self) -> UserIdentity:
        return self._userIdentity

    @property
    def InstanceIdentity(self) -> InstanceIdentity:
        return self._instanceIdentity

    @property
    def strategy(self) -> ControlStrategy:
        return self._strategy

    def __init__(self, product_id: UUID, version: str, public_key: str, user_identity: UserIdentity,
                 instance_identity: InstanceIdentity, strategy: ControlStrategy = None,
                 saved_state: ValidationState = None, session_state: ValidatedSessionState = None):
        self._state = ValidationState.fromOther(saved_state, None)
        self._sessionState = fromOtherValidatedSessionState(session_state)
        self._productId = product_id
        self._version = version
        self._userIdentity = user_identity
        self._instanceIdentity = instance_identity
        self._strategy = strategy or ControlStrategy()
        self._strategy.Lock()
        self._webService = LicensingWebService(public_key)
        self._sessionManager = None

    def Register(self, license_key: str, user_info: UserInfo) -> RegistrationResult:
        user_info.Verify()
        return ControlManagerHelper.Register(self._state, lambda: self.RegisterCore(license_key, user_info))

    def RegisterCore(self, license_key: str, user_info: UserInfo) -> RegisterResultBase:
        return self._webService.RegisterAppAsync(self._productId, license_key, user_info, self._instanceIdentity,
                                                 self._userIdentity)

    def ValidateSession(self, requested_session_duration: Optional[timedelta] = None) -> SessionValidity:
        ControlManagerHelper.InitializeState(self._state)
        if not self._state.Registered:
            return ControlManagerHelper.GetUnregisteredValidity(self._strategy, self._state,
                                                                TimeProvider.Current().UtcNow())
        self.EnsureWebSessionManagerCreated()
        session_state = self._sessionState.SessionState
        if session_state is SessionState.NotStarted:
            return self._sessionManager.BeginSession(requested_session_duration)
        elif session_state is SessionState.Ended:
            raise ValueError('The current session has already ended')
        else:
            return self._sessionManager.ExtendSession(requested_session_duration)

    def EnsureWebSessionManagerCreated(self):
        if self._sessionManager is None:
            self._sessionManager = AppSessionManager(self._webService, self._productId, self._version, self._state,
                                                     self._sessionState, self._instanceIdentity, self._userIdentity)

    def EndSession(self) -> bool:
        if self._sessionManager is None or \
                self._sessionState.SessionState is SessionState.NotStarted or \
                self._sessionState.SessionState is SessionState.Ended:
            return True
        return self._sessionManager.EndSession()

    def CheckOutComponents(self, component_names: List[str]) -> ComponentCheckoutResult:
        self.EnsureRegistered()
        self.EnsureSessionExists()
        if component_names is None or len(component_names) == 0:
            raise ValueError('At least one component must be specified')
        return self._sessionManager.CheckOutComponents(component_names)

    def CheckOutConsumableComponent(self, component_name: str, token_count: Optional[int] = None) -> ComponentCheckoutResult:
        self.EnsureRegistered()
        self.EnsureSessionExists()
        if component_name is None or len(component_name) == 0:
            raise ValueError('Component name must be specified')
        return self._sessionManager.CheckOutConsumableComponent(component_name, token_count)

    def EnsureSessionExists(self):
        if self._sessionManager is None:
            raise ValueError('A session is not active')

    def EnsureRegistered(self):
        if not self._state.Registered:
            raise ValueError('The instance must be registered')

    def GetComponentsStatus(self):
        self.EnsureRegistered()
        self.EnsureSessionExists()
        return self._sessionManager.GetComponentsStatus()

    def GetUserLicense(self) -> UserLicenseResult:
        result = self._webService.GetUserLicense(self._productId, self._userIdentity)
        return UserLicenseResult(result.Success, result.ErrorMessage, result.LicenseKeys)



import requests
from datetime import timedelta


class ConnectionSettings(object):
    DEFAULT_HOST: str = "https://elm.evoleap.com"
    DEFAULT_REQUEST_TIMEOUT: timedelta = timedelta(seconds=60)
    DEFAULT_SSL_VERIFY: bool = True

    __handler: requests
    __host: str = DEFAULT_HOST
    __request_timeout: timedelta = DEFAULT_REQUEST_TIMEOUT
    __ssl_verify: bool = DEFAULT_SSL_VERIFY

    @classmethod
    def Host(cls) -> str:
        return cls.__host

    @classmethod
    def SetHost(cls, value):
        cls.__host = value

    @classmethod
    def RequestTimeout(cls) -> timedelta:
        return cls.__request_timeout

    @classmethod
    def SetRequestTimeout(cls, timeout: timedelta):
        cls.__request_timeout = timeout

    @classmethod
    def SSLVerify(cls) -> bool:
        return cls.__ssl_verify
    
    @classmethod
    def SetSSLVerify(cls, value: bool):
        cls.__ssl_verify = value